import asyncio
import websockets
import ssl
import pathlib
import certifi
import json
import traceback
import AIArena.games as games
import os


async def handler(ai, id, game):
    uri = "wss://api.aiarena.app:80"
    hostOverride=False
    if 'AIAAPI' in os.environ and os.environ['AIAAPI'] == "localhost":
        uri = "wss://%s:9000" % os.environ['AIAAPI'] #"wss://localhost:80"
        hostOverride=True

    ssl_context=None
    
    if not hostOverride:
        ssl_context = ssl.create_default_context()
        ssl_context.load_verify_locations(certifi.where())
    
    async with websockets.connect(uri, ssl=ssl_context) as websocket:
        #send hello with array of AI IDs
        data = {
            "status":"hello",
            "payload":{
                "aiIDs":[(id, game)]
            }
        }
        await websocket.send(json.dumps(data))

        #listen for updates
        try:
            async for message in websocket:
                data = json.loads(message)
                status = data["status"]
                payload = data["payload"]
                #print(status, payload)

                #switch response based on status
                if status == "state":
                    gameID = payload["gameID"]
                    state = payload["state"]
                    aiID = payload["aiID"]

                    if game == "SpaceFighter":
                        state["cnsts"] = games.SpaceFighter_cnsts

                    #make move
                    move = ai.makeMove(state)

                    response = {
                        "status":"move",
                        "payload":{
                            "gameID":gameID,
                            "aiID":aiID,
                            "move":move
                        }
                    }
                    #send move
                    await websocket.send(json.dumps(response))

                elif status == "message":
                    print(f'{payload}')

                elif status == "game created":
                    gameID = payload["gameID"]
                    gameType = payload["gameType"]
                    aiID = payload["aiID"]
                    print(f'A new game was created. Type:{gameType}, AI:{aiID}')

                elif status == "game deleted":
                    gameID = payload["gameID"]
                    result = payload["result"]
                    aiID = payload["aiID"]
                    print(f'A game completed. Result:{result}')
        except Exception as e:
            print("Server disconnected")
            t=traceback.format_exc()
            print(t)
            #print(str(e))
        finally:
            pass



def enterMatchMaker(ai, id, game):
    asyncio.get_event_loop().run_until_complete(handler(ai, id, game))



import AIArena.games as games
import copy

class Ref:
    """ This class coordinates, executes, and determines which AI or player won a given (locally run) game. It is reasonable to assume that AIOlympics uses a similar process to perform game execution for the purposes of AI development.
    """
    
    #def __init__(self, state=None):
    def createGame(self,game,players):
        """ Initializes a given game with a given set of players (human and/or AI). This takes the place of a constructor.
        
        :param game: Name of a game to set up (currently, only "Connect4" is implemented)
        :type game: string
        :param players: List of :class:`aiolympics.AI.Player` objects to have play the game.
        :type players: list
        """
        self.pids = [x.name for x in players]
        self.players = players
        self.moves = []
        self.states = []
        self.gameName = game

        if game == "Connect4":
            self.game = games.Connect4(None, self.pids)
        
        elif game=="GymBlackjack":
            self.game=games.GymBlackjack(None, self.pids)
        
        elif game=="GymFrozenLake":
            self.game=games.GymFrozenLake(None, self.pids)

        elif game=="Arena":
            self.game=games.Arena(None, self.pids)
        else:
            print("Invalid game name, please check for typos")
            exit(-1)



    def runGame(self, display=False):
        """ Steps through player moves (in order) until the game specified by createGame() is finished.
        
        :param display: Whether or not to print every move and game state as they occur. Default value is False
        :type display: bool
        :return: The final game state. 
        :rtype: dict
        """
        
        # Treat Arena with kid gloves (sorry! We still need to sort this!)
        self.states.append(copy.deepcopy(self.game.state))
        if self.gameName=="Arena":
            print("Running Arena")
            #print(self.game.state)
            while "Winner" not in self.game.state:
                #print("Turn: %d" % self.game.state["turn"])
                for p in self.players:
                    # playerArr has the player-accessible player state:
                    if p.name in self.game.playerArr:
                        move=p.makeMove(self.game.playerArr[p.name])
                        self.moves.append(move)
                        if not self.move(move, p):
                            print("player %s made invalid move" % p.name)
                            print("Move: ", move)
                            pass
                            #return -1
                        else:
                            #print("player %s made move" % p.name)
                            #print("Move: ", move)
                            self.states.append(copy.deepcopy(self.game.state))
                    elif display:
                        #print("Skipping player "+p.name)
                        pass
                if display:
                    self.game.print()
        else:
            while "Winner" not in self.game.state:
                #print("turn", self.game.state["turn"])
                player = self.players[self.game.state["turn"]]
                move = player.makeMove(self.game.state)
                self.moves.append(move)
                if not self.move(move, player):
                    return -1
                self.states.append(copy.deepcopy(self.game.state))
                if display:
                    self.game.print()
            print("The winner is:",self.game.state["Winner"])
        return self.game.state["Winner"]


    def move(self, move, player):
        """ Runs the move passed.
        
        :param move: The move to directly run in-game.
        :return: Whether or not the move was successful.
        :rtype: bool
        """

        if self.game.validateMove(move, player):
            try:
                self.game.makeMove(move, player)
            except Exception:
                return False
            self.game.postMove()
            return True
        return False

    def endGame(self):
        pass






# Arena.py -- Implements a multiplayer arena style deathmatch between AI agents.
# General operation:
# Arena.py works on a simple 50x50 grid where each player may take up one space.
# On each "turn", an agent may opt to do one of the following:
#  1. Move one space in any direction
#  2. Pick up an object present on the space
#  3. "Fire" on a given space with coordinates (depending on the object/weapon equipped)
# Each space is represented as a tuple with the following attrributes:
# (player, item)
# player is either None or the name of the AI on that space.
# item is either None or an int referring to one of the following item codes:
# 0 - health pickup (+10 health)
# 1 - AoE pickup (increases the radius of attacks by one tile)
# TODO: add more of these
# 
# On each turn, the agent will be passed the following game state information:
# {"health": player health as an integer,
#  "board": game board information,
#  "position": the player's position (tuple; row, column)
#  "inventory": list of items picked up
#  "range": The player's firing range (max euclidean distance to a given tile)
#  "aoe": The radius (in tiles) of splash damage on the player's shots
#  "lastmove": bool representing whether the last move was valid
#  "score": Integer representing the player's current score.
# }
# TODO: Implement items and such. So far they will be stripped just to get the game working quickly.

import random
import math
import copy

ACTIONS=[None, "move", "none", "pickup", "fire"]
ITEMS=[{"aoe": 2, "range": -1, "health": 0}, # Area of Effect pickup
       {"aoe": 0, "range": 3, "health": 0}, # Range pickup
       {"aoe": 0, "range": 0, "health": 25}, # Health pickup
       {"aoe": random.randint(-1, 2), "range": random.randint(-5, 5), "health": random.randint(-10, 30)} # Random bonus pickup.
       ]
BOARDWIDTH=25
BOARDHEIGHT=25

def place(board, player, pos, oldPos=None):
    if board[pos[0]][pos[1]][0] is None:
        # Update the board tuple here with (player, oldcontents)
        board[pos[0]][pos[1]]=(player, board[pos[0]][pos[1]][1])
        
        if oldPos is not None:
            board[oldPos[0]][oldPos[1]]=(None, board[oldPos[0]][oldPos[1]][1])
        return True
    return False

def placeItem(board, item, pos, oldPos=None):
    if board[pos[0]][pos[1]][1] is None:
        board[pos[0]][pos[1]]=(board[pos[0]][pos[1]][0], item)
        
        if oldPos is not None:
            board[oldPos[0]][oldPos[1]]=(board[oldPos[0]][oldPos[1]][0], None)
        return True
    return False

def euDist(a, b):
    return math.sqrt((a[0]-b[0])**2+(a[1]-b[1])**2)

def checkMove(oldPos, newPos, maxDist=2):
    # The player 
    #print(euDist(oldPos, newPos), oldPos, newPos)
    if newPos[0]<0 or newPos[1]<0 or newPos[0]>=BOARDHEIGHT or newPos[1]>=BOARDWIDTH:
        return False
    
    # The player may move one tile in any direction per turn:
    else:
        # This calculates the overall movement between the two spots:
        dist=euDist(newPos, oldPos)
        # We allow movement along diagonals:
        return dist<maxDist

def doFire(board, target, aoe, baseDamage=10):
    playerDamage={}
    for i, cols in enumerate(board):
        for j, tile in enumerate(cols):
            bPos=(i, j)
            dist=euDist(bPos, target)
            # This means we're within the AoE to actually hit something:
            if dist<=aoe:
                player=board[i][j][0]
                if player is not None:
                    # uses dist+1 so that we don't break when dist is actually 0
                    playerDamage[player]=int(baseDamage/(dist+1))
    return playerDamage


def genArenaPlayer(name, pos, health, aoe, rng, inventory=[], lastmove=None, score=0, board=None):
    arPlayer={}
    arPlayer["name"]=name
    arPlayer["position"]=pos
    arPlayer["health"]=health
    arPlayer["aoe"]=aoe
    arPlayer["lastmove"]=lastmove
    arPlayer["score"]=score
    arPlayer["inventory"]=inventory
    arPlayer["board"]=board
    arPlayer["range"]=rng
    return arPlayer


class Arena:
    def __init__(self, state=None, players=None, replayStates=None, moves=None):
        #Game Constants
        #Moves are spec'd in row, col
        self.dims = (BOARDHEIGHT, BOARDWIDTH)
        
        self.board=[[(None, None) for i in range(self.dims[1])] for j in range(self.dims[0])]

        self.models = {}
        for ai in players:
            self.models[ai.id] = ai
        print(players)
        players = [x.id for x in players]
        self.players=players
        print(players)
        self.originalPlayers=players.copy()
        
        # randomly seed the board with items:
        for i in ITEMS:
            pos=(random.randint(0, self.dims[0]-1), random.randint(0, self.dims[1]-1))
            while not placeItem(self.board, i, pos):
                pos=(random.randint(0, self.dims[0]-1), random.randint(0, self.dims[1]-1))
                
        for i in range(self.dims[0]):
            for j in range(self.dims[1]):
                pass
        
        # randomly add players to the board:
        self.playerPos={}
        for p in players:
            pos=(random.randint(0, self.dims[0]-1), random.randint(0, self.dims[1]-1))
            while not place(self.board, p, pos):
                pos=(random.randint(0, self.dims[0]-1), random.randint(0, self.dims[1]-1))
            
            # pos should now be the player's position:
            self.playerPos[p]=pos
        
        # Let's do some dictionary comprehension, lads:
        self.playerArr={p: genArenaPlayer(p, self.playerPos[p], 100, 2, 10, board=self.board) for p in players}
        self.lastMove=None
        
        self.shortMoveList=[]
        
        if state:
            self.state = state
        else:
            self.state = {
                "turn": 0,
                "players": players,
                "board": self.board,
                "playerpos": self.playerPos,
                "players": self.playerArr
            }
        if replayStates:
            self.replayStates = replayStates
        else:
            self.replayStates = []
        if moves:
            self.moves = moves
        else:
            self.moves = []


    def runGame(self, display=False):
        result = {
            "error": "none",
            "replayStates": self.replayStates,
            "moves": self.moves
        }

        print("Running Arena")
        # print(self.game.state)
        while "Winner" not in self.state:
            print("Turn: %d" % self.state["turn"])
            for p in self.players:
                # playerArr has the player-accessible player state:
                #Fprint(p, self.playerArr)
                if p in self.playerArr:
                    move = self.models[p].makeMove(self.playerArr[p])
                    self.moves.append(move)
                    if not self.makeMove(move, self.models[p]):
                        pass
                        """print("player %s made invalid move" % p)
                        print("Move: ", move)
                        result['rewards'] = self.genErrRewards(p)
                        result['error'] = "AI : " + p + " made an illegal move"
                        return result"""
                        # return -1
                    else:
                        print("player %s made move" % p)
                        print("Move: ", move)
                        self.replayStates.append(copy.deepcopy(self.state))
                elif display:
                    # print("Skipping player "+p.name)
                    pass
            self.postMove()
            if display:
                self.print()

        result['rewards'] = self.genRewards(self.state["Winner"])
        return result

    def genRewards(self, winner):
        rewards = []
        for i, player in enumerate(self.players):
            if i == winner:
                r = 1
            else:
                r = -1
            rewards.append((player, r))
        return rewards

    def genErrRewards(self, screwUp):
        rewards = []
        for i, player in enumerate(self.players):
            if i == screwUp:
                r = -1
            else:
                r = 0
            rewards.append((player, r))
        return rewards

    def colorFormat(self, plrName, text):
        color=41+self.originalPlayers.index(plrName)
        return "\033[1;%dm%s\033[0m" % (color, text)
        
    def print(self):
        print("Round %d---------------" % self.state["turn"])
        for i, cols in enumerate(self.board):
            if i==0:
                trStr="*"
                for j in range(len(cols)):
                    trStr+=str(j%10)
                print(trStr)
            rStr=""
            for j, tile in enumerate(cols):
                if j==0:
                    rStr+=str(i%10)
                if tile[0] is None and tile[1] is None:
                    rStr+=" "
                elif tile[1] is not None:
                    rStr+="I"
                else:
                    #plrColor=self.originalPlayers.index(tile[0])
                    rStr+=self.colorFormat(tile[0], "P")#"P"
            print(rStr)
        print()
        print("Moves this turn: ")#+str(self.lastMove))
        for m in self.shortMoveList:
            print(self.colorFormat(m[0], m[0]), m[1])
        print("--------------------")
        print()

    def validateMove(self, move, player):
        player = player.id
        #print("Attempting to validate move ", move, " from player ", player)
        try:
            pl=None
            if player in self.playerArr:
                pl=self.playerArr[player]
            
            # Determine if the player is dead:
            if pl is None:
                #print("pl was none")
                return False
            
            # Determine if the move is actually valid (these will throw an exception if the move
            # is improperly formatted)
            action=move["action"]
            params=move["params"]
            maxDist=pl["range"]
            
            # Now validate the action type:
            if action in ACTIONS:
                if action is None or action=="none":
                    return True
                elif action=="move":
                    return checkMove(pl["position"], params)
                elif action=="fire":
                    #print("player is attempting to fire")
                    # once again, if this fails, then we automatically return false.
                    return checkMove(pl["position"], params, maxDist)
                elif action=="pickup":
                    return True # No parameters needed, really.
                # Go ahead and specify whatever you want, player. "pickup" doesn't take an argument, and
                # invalid commands won't even be executed.
                else:
                    return True
            else:
                pass
                #print("action not valid")
            return False
            
        except:
            return False

    def makeMove(self, move, player):#assumes move has been validated
        player = player.id
        # Step 1: determine what the player actually wants:
        action=move["action"]
        params=move["params"]
        
        if player not in self.playerArr:
            return
        
        if action=="move":
            newPos=params
            oldPos=self.playerPos[player]
            # Hoo boy...
            self.playerPos[player]=newPos
            self.playerArr[player]["position"]=newPos
            # Now we need to actually update the board:
            place(self.board, player, newPos, oldPos)
        
        elif action=="fire":
            # TODO: add player base damage (rather than the default 10:
            playerDmg=doFire(self.board, params, self.playerArr[player]["aoe"])
            # Update all players who got hit:
            for p in playerDmg:
                pl=self.playerArr[p]
                pl["health"]-=playerDmg[p]
                if pl["health"]<=0:
                    print("Player %s died!" % self.colorFormat(p, p))
                    # Remove them from the game:
                    plPos=self.playerPos[p]
                    self.board[plPos[0]][plPos[1]]=(None, self.board[plPos[0]][plPos[1]][1])
                    del self.playerPos[p]
                    del self.playerArr[p]
        
        elif action=="pickup":
            # We don't actually do anything here (yet!)
            playerPos=self.playerPos[player]
            item=self.board[playerPos[0]][playerPos[1]][1]
            if item is not None:
                self.playerArr[player]["aoe"]+=item["aoe"]
                self.playerArr[player]["range"]+=item["range"]
                self.playerArr[player]["health"]+=item["health"]
                self.board[playerPos[0]][playerPos[1]]=(self.board[playerPos[0]][playerPos[1]][0], None)

            pass
        self.lastMove=(player, move)
        if len(self.shortMoveList)>0 and len(list(self.playerArr.keys()))>0:
            if self.shortMoveList[-1][0]==list(self.playerArr.keys())[-1]:
                self.shortMoveList=[]
        self.shortMoveList.append(self.lastMove)
        return True

    # perform cleanup and determine if the game is over:
    def postMove(self):
        if len(self.playerArr)==1:
            self.endGame(list(self.playerArr)[0])
        elif len(self.playerArr)==0:
            self.endGame(None)
        else:
            # This just indicates that there has been another timestep.
            # The ref keeps track of move history so we don't have to (which is very convenient!)
            self.state["turn"]=self.state["turn"]+1

    def endGame(self, winner):
        #print("Winner", winner)
        #print(self.state["players"][winner])
        winnerInd = self.originalPlayers.index(winner)
        self.state["Winner"] = winnerInd
        self.shortMoveList=[(winner, 'won!')]



import dill as pickle
import requests

def upload(Ai, game=None, aid=None):
    if aid is None:
        print("Please paste in the AI Id as the third parameter")
        return
    if game is None:
        print("Please enter the game as the second parameter")
        return

    payload = {
        "Ai":Ai,
        "aid":aid,
        "game":game
    }
    sPayload = pickle.dumps(payload,recurse=True)
    r = requests.post("https://us-central1-ai-olympics.cloudfunctions.net/saveAI",sPayload)
    print("status code:",r.status_code)
    print("response:",r.content.decode())





import copy
import random

cnsts = {}

#board
cnsts['boardWidth'] = 100
cnsts['boardHeight'] = 200

#ships
cnsts['shipDim'] = 10
shipDim_h = cnsts['shipDim'] / 2
cnsts['shipVelX'] = 2.5
cnsts['shipVelY'] = 0.02
cnsts['shipHealth'] = 1000
cnsts['shipReload'] = 30


cnsts['minXpos'] = cnsts['shipDim']
cnsts['maxXpos'] = cnsts['boardWidth'] - cnsts['shipDim']

#bullets
cnsts['bulletDim'] = 3
cnsts['bulletSpeed'] = 3
cnsts['bulletDamage'] = 30
cnsts['bulletsTTL'] = 5

bulletDim_h = cnsts['bulletDim'] / 2

#items
cnsts['ticksPerItem'] = 100
cnsts['itemDim'] = 10
itemDim_h = cnsts['itemDim'] / 2
cnsts['itemSpeed'] = 1.5

class SpaceFighter:
    def __init__(self, state=None, remotePlayers=None):
        self.aiIDs = remotePlayers
        self.flippedAI = self.aiIDs[1]
        self.playerViews = {}

        if state == None:
            self.state = {
                "ships": {
                    self.aiIDs[0]: {
                        "pos": [cnsts['shipDim'],cnsts['shipDim']],
                        "health": cnsts['shipHealth'],
                        "velocity": [0, cnsts['shipVelY']],
                        "reload": 0,
                        "target": [cnsts['boardWidth'] / 2 , cnsts['boardHeight'] / 2],
                        "level": 0
                    },
                    self.aiIDs[1] : {
                        "pos": [cnsts['boardWidth']-cnsts['shipDim'], cnsts['boardHeight']-cnsts['shipDim']],
                        "health": cnsts['shipHealth'],
                        "velocity": [0, -cnsts['shipVelY']],
                        "reload": 0,
                        "target": [cnsts['boardWidth'] / 2 , cnsts['boardHeight'] / 2],
                        "level": 0
                    }
                },

                "objects": {},
                "objCnt":0,
                "tickCnt":0
            }
        else:
            self.state = state

        self.tick()
        print(self.playerViews)

    def getStateForPlayer(self,aiID):
        return self.playerViews[aiID]

    # move:{"move":"left/right","fire":True/False
    def validateMove(self,move,playerID=None):
        if "move" not in move and "target" not in move:
            return False

        if ("move" in move and not (move["move"] == "left" or move["move"] == "right")):
            return False

        if ("target" in move and not (type(move["target"]) == list) and not (len(move["target"]) == 2)):
            return False

        return True

    def makeMove(self, move, playerID=None):
        aiID = playerID
        if "move" in move:
            direction = move["move"]
            if aiID == self.flippedAI:
                direction = "left" if direction == "right" else "right"


            dx = cnsts['shipVelX']
            if direction == "left":
                dx = -dx
            self.state["ships"][aiID]["velocity"][0] = dx

            newX = self.state["ships"][aiID]["pos"][0] + dx
            if newX < cnsts['minXpos']:
                newX = cnsts['minXpos']
            if newX > cnsts['maxXpos']:
                newX = cnsts['maxXpos']

            self.state["ships"][aiID]["pos"][0] = newX
        else:
            self.state["ships"][aiID]["velocity"][0] = 0

        if "target" in move:
            target = move["target"]
            if aiID == self.flippedAI:
                target = self.flipCoords(target)

            if aiID != self.flippedAI:
                minY = self.state["ships"][aiID]['pos'][1] + shipDim_h + bulletDim_h + 1
                if target[1] < minY:
                    target[1] = minY
            else:
                maxY = self.state["ships"][aiID]['pos'][1] - shipDim_h - bulletDim_h - 1
                if target[1] > maxY:
                    target[1] = maxY

            self.state["ships"][aiID]["target"] = target


    def newObjectID(self):
        objId = "obj_" + str(self.state["objCnt"])
        self.state["objCnt"] += 1
        return objId

    def newBullet(self, pos, velocity, color):
        objId = self.newObjectID()

        self.state["objects"][objId] = {
            "type": "bullet",
            "pos": pos,
            "velocity": velocity,
            "color": color,
            "ttl": cnsts['bulletsTTL']
        }

    def newExplosion(self, pos):
        objId = self.newObjectID()

        self.state["objects"][objId] = {
            "type": "explosion",
            "pos": pos,
            "ttl": 20
        }

    def newItem(self):
        objId = self.newObjectID()

        side = 1 if random.random() < 0.5 else -1
        posX = 50 + (side * (50 + (random.random() * 50)))
        posY = (random.random() * 100) + 50
        velX = random.random() * cnsts['itemSpeed'] * -side
        velY = (random.random() - 0.5) *cnsts['itemSpeed']

        self.state["objects"][objId] = {
            "type": "item",
            "pos": [posX,posY],
            "velocity": [velX,velY],
        }

    def tick(self):
        self.state["tickCnt"] += 1

        #update reloads and autofire and move closer
        for key, object in self.state["ships"].items():
            self.state["ships"][key]["reload"] = object["reload"] - 1
            if object["reload"] <= 1:
                posX = self.state["ships"][key]["pos"][0]
                posY = self.state["ships"][key]["pos"][1] + shipDim_h + bulletDim_h + 1

                if key == self.flippedAI:
                    posY = self.state["ships"][key]["pos"][1] - shipDim_h - bulletDim_h - 1


                velX = object['target'][0] - posX  # dx
                velY = object['target'][1] - posY  # dy
                uVec = self.toUnitVector([velX, velY])
                velX = uVec[0] * cnsts['bulletSpeed']  # unit vector then speed
                velY = uVec[1] * cnsts['bulletSpeed']  # unit vector then speed

                color = "red" if key == self.flippedAI else "blue"
                self.newBullet([posX, posY], [velX, velY], color)

                newReload = max(5, cnsts['shipReload'] - (2 * self.state["ships"][key]['level']))
                self.state["ships"][key]["reload"] = newReload

            self.state["ships"][key]['pos'][1] = object['pos'][1] + object['velocity'][1]


        #update objects
        deletes = []
        for key, object in self.state["objects"].items():
            if object["type"] == "bullet" or object["type"] == "item":
                curPos = self.state["objects"][key]["pos"]
                velocity = self.state["objects"][key]["velocity"]
                self.state["objects"][key]["pos"] = [curPos[0] + velocity[0], curPos[1] + velocity[1]]
            elif object["type"] == "explosion":
                if object["ttl"] <= 0:
                    deletes.append(key)
                self.state["objects"][key]["ttl"] = object["ttl"] - 1


        #check collisions
        #ships
        collisions = self.getShipCollisions()
        for aiID, objID in collisions:
            if self.state["objects"][objID]["type"] == "bullet":
                self.state["ships"][aiID]["health"] -= cnsts['bulletDamage']
                deletes.append(objID)
                self.newExplosion(self.state["ships"][aiID]['pos'])

        #objects
        collisions = self.getObjectCollisions()
        for objAID, objBID in collisions:
            if self.state["objects"][objAID]["type"] == "bullet" and self.state["objects"][objBID]["type"] == "bullet":
                self.newExplosion(self.state["objects"][objAID]["pos"])
                deletes.append(objAID)
                deletes.append(objBID)
            elif self.state["objects"][objAID]["type"] == "item" or self.state["objects"][objBID]["type"] == "item":
                itemID = objAID
                bullID = objBID
                if self.state["objects"][objBID]["type"] == "item":
                    itemID = objBID
                    bullID = objAID

                shipID = self.aiIDs[0]
                if self.state["objects"][bullID]["color"] == "red":
                    shipID = self.aiIDs[1]
                self.state['ships'][shipID]['level'] += 1

                self.newExplosion(self.state["objects"][objAID]["pos"])
                deletes.append(objAID)
                deletes.append(objBID)





        #check bounds of objects
        for key, object in self.state["objects"].items():
            if object["type"] == "bullet":
                if object["pos"][1] > cnsts['boardHeight'] or object["pos"][1] < 0:
                    deletes.append(key)
                if object["pos"][0] > cnsts['boardWidth'] or object["pos"][0] < 0:
                    curVel = self.state["objects"][key]["velocity"]
                    self.state["objects"][key]["velocity"][0] = -curVel[0]

                    curTTL = self.state["objects"][key]["ttl"]
                    curTTL -= 1
                    self.state["objects"][key]["ttl"] = curTTL
                    if curTTL <= 0:
                        deletes.append(key)
            elif object["type"] == "item":
                curVel = self.state["objects"][key]["velocity"]

                if object["pos"][0] < 0:
                    self.state["objects"][key]["velocity"][0] = abs(curVel[0])
                if object["pos"][0] > cnsts['boardWidth']:
                    self.state["objects"][key]["velocity"][0] = -abs(curVel[0])

                if object["pos"][1] < 0:
                    self.state["objects"][key]["velocity"][1] = abs(curVel[1])
                if object["pos"][1] > cnsts['boardHeight']:
                    self.state["objects"][key]["velocity"][1] = -abs(curVel[1])

        deletes = list(set([i for i in deletes]))
        for key in deletes:
            del self.state["objects"][key]

        #check health
        if self.state["ships"][self.aiIDs[0]]["health"] <= 0:
            self.state["Winner"] = 1
        elif self.state["ships"][self.aiIDs[1]]["health"] <= 0:
            self.state["Winner"] = 0


        #item creation
        if self.state["tickCnt"] % cnsts['ticksPerItem'] == 0:
            self.newItem()

        #update views
        self.generateViews()
        #print(self.playerViews)

    def getShipCollisions(self):
        collisions = []
        for key, object in self.state["objects"].items():
            if object["type"] == "bullet":
                if self.isColliding(self.state["ships"][self.aiIDs[0]]["pos"], "ship", object["pos"], object["type"]):
                    collisions.append((self.aiIDs[0], key))
                if self.isColliding(self.state["ships"][self.aiIDs[1]]["pos"], "ship", object["pos"], object["type"]):
                    collisions.append((self.aiIDs[1], key))
        return collisions

    def getObjectCollisions(self):
        collisions = []
        for key, object in self.state["objects"].items():
            if object["type"] == "bullet":
                for keyB, objectB in self.state["objects"].items():
                    if key == keyB:
                        continue
                    if objectB["type"] == "bullet" or objectB["type"] == "item":
                        if self.isColliding(object["pos"], object["type"], objectB["pos"], objectB["type"]):
                            if key < keyB:
                                collisions.append((key, keyB))
                            else:
                                collisions.append((keyB, key))

        #remove duplicates
        return list(set([i for i in collisions]))

    def isColliding(self, aPos, aType, bPos, bType):
        aDim_h = self.typeToDimH(aType)
        bDim_h = self.typeToDimH(bType)

        if aType == "bullet" and bType == "bullet":
            return self.isColliding_Circle_Circle(aPos, aDim_h, bPos, bDim_h)

        elif aType == "ship" or bType == "ship":
            shipPos = aPos
            bullPos = bPos
            if bType == "ship":
                shipPos = bPos
                bullPos = aPos

            return self.isColliding_Circle_Box(bullPos, bulletDim_h, shipPos, shipDim_h)

        elif aType == "item" or bType == "item":
            itemPos = aPos
            bullPos = bPos
            if bType == "item":
                itemPos = bPos
                bullPos = aPos

            return self.isColliding_Circle_Circle(bullPos, bulletDim_h, itemPos, itemDim_h)

    def isColliding_Circle_Circle(self, aPos, aDim_h, bPos, bDim_h):
        dx = aPos[0] - bPos[0]
        dy = aPos[1] - bPos[1]
        distance = (dx ** 2 + dy ** 2) ** (1 / 2)
        if distance < aDim_h + bDim_h:
            return True
        return False

    def isColliding_Box_Box(self, aPos, aDim_h, bPos, bDim_h):
        overlapX = True if (abs(aPos[0] - bPos[0]) < (aDim_h + bDim_h)) else False
        overlapY = True if (abs(aPos[1] - bPos[1]) < (aDim_h + bDim_h)) else False
        if overlapX and overlapY:
            return True
        return False

    def isColliding_Circle_Box(self, cPos, cDim_h, bPos, bDim_h):
        dx = abs(cPos[0] - bPos[0])
        dy = abs(cPos[1] - bPos[1])

        if dx > (cDim_h + bDim_h) or dy > (cDim_h + bDim_h): #centers too far
            return False

        if dx <= (cDim_h + bDim_h) or dy <= (cDim_h + bDim_h):#centers must be close enough
            return True

        cornerDist = (dx**2 + dy**2)**(1/2) #last edge case of corners overlapping
        if cornerDist < cDim_h + bDim_h:
            return True

        return False


    def typeToDimH(self, type):
        if type == "bullet":
            return bulletDim_h
        elif type == "ship":
            return shipDim_h

    def generateViews(self):
        #standard view
        self.playerViews[self.aiIDs[0]] = {
            "me": copy.deepcopy(self.state["ships"][self.aiIDs[0]]),
            "enemy": copy.deepcopy(self.state["ships"][self.aiIDs[1]]),
            "objects": copy.deepcopy(self.state["objects"])
        }

        #flipped view
        flippedView = {
            "me": copy.deepcopy(self.state["ships"][self.aiIDs[1]]),
            "enemy": copy.deepcopy(self.state["ships"][self.aiIDs[0]]),
            "objects": copy.deepcopy(self.state["objects"])
        }

        flippedView["me"]["pos"] = self.flipCoords(flippedView["me"]["pos"])
        flippedView["enemy"]["pos"] = self.flipCoords(flippedView["enemy"]["pos"])

        for key, object in flippedView["objects"].items():
            flippedView["objects"][key]["pos"] = self.flipCoords(object["pos"])
            if object["type"] == "bullet":
                flippedView["objects"][key]["velocity"] = [-object["velocity"][0], -object["velocity"][1]]
                flippedView["objects"][key]["color"] =  "red" if object["color"] == "blue" else "blue"

        self.playerViews[self.aiIDs[1]] = flippedView


    def flipCoords(self, coors):
        return [cnsts['boardWidth'] - coors[0], cnsts['boardHeight'] - coors[1]]

    def toUnitVector(self, vec):
        magnitude = (vec[0] ** 2 + vec[1] ** 2) ** (1 / 2)
        return [vec[0] / magnitude, vec[1] / magnitude]


if __name__ == "__main__":
    game = SpaceFighter(remotePlayers=["alice","bob"])



import random
import copy

class Ants:
    def __init__(self, state=None, AIs=None):
        self.bounds = (10,10)

        self.AIs = AIs
        if state is not None:
            self.state = state
        else:
            self.freshState()


    def freshState(self):
        self.state = {"bots":[]}
        usedPos = []
        for i in range(len(self.AIs)):
            for _ in range(10): #Soldier
                newPos = (random.randint(0,self.bounds[0]-1), random.randint(0,self.bounds[1]-1))
                while newPos in usedPos:
                    newPos = (random.randint(0,self.bounds[0]-1), random.randint(0,self.bounds[1]-1))
                usedPos.append(newPos)
                bot = {
                    "ai":i,
                    "pos":newPos,
                    "health":100,
                    "type":"Soldier"
                }
                self.state["bots"].append(bot)

            for _ in range(2): #Queen
                newPos = (random.randint(0,self.bounds[0]-1), random.randint(0,self.bounds[1]-1))
                while newPos in usedPos:
                    newPos = (random.randint(0,self.bounds[0]-1), random.randint(0,self.bounds[1]-1))
                usedPos.append(newPos)
                bot = {
                    "ai":i,
                    "pos":newPos,
                    "health":100,
                    "type":"Queen",
                    "breedWait":3
                }
                self.state["bots"].append(bot)


    def printBoard(self):
        board = []
        for i in range(self.bounds[0]):
            col = []
            for j in range(self.bounds[1]):
                col.append(0)
            board.append(col)

        for bot in self.state["bots"]:
            board[bot["pos"][0]][bot["pos"][1]] = (bot["type"][0], bot["ai"])

        pStr = "= " * (self.bounds[0] + 2)
        print(pStr)
        for j in range(self.bounds[1]):
            rj = self.bounds[1] - 1 - j
            pStr = "= "
            for i in range(self.bounds[0]):
                if board[i][rj] == 0:
                    pStr += "  "
                else:
                    pStr += self.colorFormat(board[i][rj][1], board[i][rj][0] + " ")
            pStr += "= "
            print(pStr)
        pStr = "= " * (self.bounds[0] + 2)
        print(pStr)

    def colorFormat(self, index, text):
        color=41+index
        return "\033[1;%dm%s\033[0m" % (color, text)

    def runGame(self):
        self.printBoard()
        result = 1
        while result == 1:
            result = self.round()
            self.printBoard()
            x = input("hit enter for next move...")

    def round(self):
        #make moves
        #print(self.state)
        for bot in self.state["bots"]:
            moveState = copy.deepcopy(self.state)
            moveState["bot"] = copy.deepcopy(bot)
            action, position = self.AIs[bot["ai"]].makeMove(moveState)
            #print(action,position)
            target = None
            for tbot in self.state["bots"]:
                if tbot["pos"] == position:
                    target = tbot

            if action == "move" and target == None:
                if position[0] >= 0 and position[0] < self.bounds[0] and position[1] >= 0 and position[1] < self.bounds[1]:
                    if abs(position[0] - bot["pos"][0]) <= 1 and abs(position[1] - bot["pos"][1]) <= 1:
                        bot["pos"] = position

            if action == "attack" and target != None:
                if abs(position[0] - bot["pos"][0]) <= 1 and abs(position[1] - bot["pos"][1]) <= 1:
                    dmg = 34
                    target["health"] = target["health"] - dmg

            if action == "breed" and bot["type"] == "Queen":
                if bot["breedWait"] == 0:
                    newBot = {
                        "ai": bot["ai"],
                        "pos": bot["pos"],
                        "health": 100,
                        "type": "Soldier"
                    }
                    self.state["bots"].append(newBot)
                    bot["breedWait"] = 3
                else:
                    bot["breedWait"] -= 1

        #check for dead to remove
        botsCnt = len(self.state["bots"])
        for i in range(botsCnt):
            if self.state["bots"][botsCnt - 1 - i]["health"] <= 0:
                del self.state["bots"][botsCnt - 1 - i]

        #check if all dead
        if len(self.state["bots"]) == 0:
            return -1

        #check for two different ais
        match = self.state["bots"][0]["ai"]
        for bot in self.state["bots"]:
            if bot["ai"] != match:
                return 1

        #winner
        return 0





class AntsAI:
    def __init__(self, id="antsAI"):
        self.id = id
        self.game = "Ants"

    def makeMove(self,state):
        if state["bot"]["type"] == "Soldier":
            move = self.makeMove_Soldier(state)
        elif state["bot"]["type"] == "Queen":
            move = self.makeMove_Queen(state)
        return move

    def makeMove_Soldier(self,state):
        closest = 99999999
        closestPos = None
        myPos = state["bot"]["pos"]
        for bot in state["bots"]:
            if bot == state["bot"]:#Me
                continue
            if bot["ai"] == state["bot"]["ai"]:#My team
                continue

            dist = ((myPos[0]-bot["pos"][0])**2 + (myPos[1]-bot["pos"][1])**2)**(1/2)
            if dist < closest:
                closest = dist
                closestPos = bot["pos"]

        #print(closest,closestPos)

        if closest < 2:
            return "attack", closestPos

        dx = 0
        if closestPos[0] > myPos[0]:
            dx = 1
        elif closestPos[0] < myPos[0]:
            dx = -1

        dy = 0
        if closestPos[1] > myPos[1]:
            dy = 1
        elif closestPos[1] < myPos[1]:
            dy = -1

        return "move", (myPos[0] + dx, myPos[1] + dy)

    def makeMove_Queen(self, state):
        return "breed", (0,0)


if __name__ == "__main__":
    ai1 = AntsAI(id="ai1")
    ai2 = AntsAI(id="ai2")
    game = Ants(state=None,AIs=[ai1,ai2])
    game.runGame()


from AIArena.games.Connect4 import Connect4 as Connect4
#from AIArena.games.GymBlackjack import GymBlackjack as GymBlackjack
#from AIArena.games.GymFrozenLake import GymFrozenLake as GymFrozenLake
from AIArena.games.Arena import Arena as Arena
from AIArena.games.StockMarket import StockMarket as StockMarket

from AIArena.games.SpaceFighter import SpaceFighter as SpaceFighter
from AIArena.games.SpaceFighter import cnsts as SpaceFighter_cnsts

from AIArena.games.BattleTanks import BattleTanks as BattleTanks



class Player:
    """This forms an interface for a generic game "player." It is extended by the AI class and can be used to represent any kind of entity that makes moves in a given game.
    
    :param name: The human-readable reported name for the player (ex. LilBuddy for our Connect 4 playing AI)
    :param game: The specific game instance for this AI to play.
    """
    def __init__(self, name, game):
        """Constructor
        """
        self.name = name
        self.game = game
        if not hasattr(self, "id"):
            self.id = self.name

    def makeMove(self, state):
        """Returns an appropriate move given a game state. Must be overridden in implementations of the Player class.
        :param state: The game state given which to make a move.
        :return: An appropriate move representation depending on the game played.
        """
        print("This player is missing the method 'makeMove(self, state)'")
        return None

class AI(Player):
    """This is the core interface for AI development. Extending the makeMove() method allows for a hook into an AI so that it can respond to game states and inputs.
    
    :param name: The human-readable reported name for this AI (ex. LilBuddy for our Connect 4 playing AI)
    :param game: The specific game instance for this AI to play.
    """
    def __init__(self, name, game):
        """Constructor. Calls superclass (Player) constructor and nothing else
        """
        super().__init__(name, game)
        #add AI specific code here

    def makeMove(self, state):
        """Returns an appropriate move given a game state. Must be overridden in implementations of the Player class.
        :param state: The game state in which to make a move.
        :return: An appropriate move representation depending on the game played.
        """
        print("Your AI is missing the method 'makeMove(self, state)'")
        return None

class AI_Connect4(AI):
    def __init__(self, name):
        super().__init__(name, "Connect4")






# BattleTanks.py -- Implements a multiplayer arena style deathmatch between AI agents.
# General operation:
# Arena.py works on a simple 50x50 grid where each player may take up one space.
# On each "turn", an agent may opt to do one of the following:
#  1. Move one space in any direction
#  2. Pick up an object present on the space
#  3. "Fire" on a given space with coordinates (depending on the object/weapon equipped)
# Each space is represented as a tuple with the following attrributes:
# (player, item)
# player is either None or the name of the AI on that space.
# item is either None or an int referring to one of the following item codes:
# 0 - health pickup (+10 health)
# 1 - AoE pickup (increases the radius of attacks by one tile)
# TODO: add more of these
# 
# On each turn, the agent will be passed the following game state information:
# {"health": player health as an integer,
#  "board": game board information,
#  "position": the player's position (tuple; row, column)
#  "inventory": list of items picked up
#  "range": The player's firing range (max euclidean distance to a given tile)
#  "aoe": The radius (in tiles) of splash damage on the player's shots
#  "lastmove": bool representing whether the last move was valid
#  "score": Integer representing the player's current score.
# }

import random
import math
import copy
import uuid

WIDTH=25*32
HEIGHT=25*32
HEALTH=100
ITEMS=[{"aoe": 2, "range": -1, "health": 0}, # Area of Effect pickup
       {"aoe": 0, "range": 3, "health": 0}, # Range pickup
       {"aoe": 0, "range": 0, "health": 25}, # Health pickup
       {"aoe": random.randint(-1, 2), "range": random.randint(-5, 5), "health": random.randint(-10, 30)} # Random bonus pickup.
       ]

PROJECTILESPLASH=32
PROJECTILEDAMAGE=30
PROJECTILESCLFAC=1
BOMBSPLASH=128
BOMBDAMAGE=20
BOMBSCLFAC=0.5
MOVESPEED=5
PROJMOVESPEED=10


# Utility functions:

def euDist(a, b):
    return math.sqrt((a[0]-b[0])**2+(a[1]-b[1])**2)

class BTObject:
    def __init__(self, id, pos, vel, health, type, inventory, width, height, timer=-1):
        self.id=id
        self.pos=pos
        self.vel=vel
        self.health=health
        self.inventory=inventory
        self.type=type
        self.width=width
        self.height=height
        self.timer=timer
        
    def getBoundingBox(self):
        # Top left, top right, bottom left bottom right
        return [(self.pos[0]-self.width/2, self.pos[1]-self.height/2), (self.pos[0]+self.width/2, self.pos[1]-self.height/2), (self.pos[0]-self.width/2, self.pos[1]+self.height/2), (self.pos[0]+self.width/2, self.pos[1]+self.height/2)]
        
    def detectHit(self, other):
        ourbb=self.getBoundingBox()
        thrbb=other.getBoundingBox()
        hit=False
        for pt in thrbb:
            hit=hit or (pt[0]>=ourbb[0][0] and pt[0]<=ourbb[3][0] and pt[1]>=ourbb[0][1] and pt[1]<=ourbb[3][1])
        
        return hit
        
    # Returns a set of objects struck + their associated damages if our type is correct
    def splash(self, otherObjects, basedmg, scalefactor, radius):
        objs=[]
        for o in otherObjects:
            if o.type!="explosion":
                # Don't perform splash for explosions (ie. graphics)
                dist=euDist(self.pos, o.pos)
                if dist<radius:
                    #objs.append((o.id, playerDamage[player]=int(basedmg/((dist+1)*scalefactor))))
                    #Tristans attempt at fixing:
                    objs.append((o.id, int(basedmg / ((dist+1) * scalefactor))))
        return objs
    
    def move(self):
        self.pos[0]+=self.vel[0]
        self.pos[1]+=self.vel[1]
        
        # This might prove to be an interesting intentional quirk: have rigid collisions with the board bounds:
        if self.pos[0]>WIDTH:
            self.vel[0]*=-1
            self.pos[0]=WIDTH
            print("BOUNCE")
        elif self.pos[0]<0:
            print("BOUNCENEG")
            self.vel[0]*=-1
            self.pos[0]=0
        
        if self.pos[1]>HEIGHT:
            print("BOUNCEY")
            self.vel[1]*=-1
            self.pos[1]=HEIGHT
        elif self.pos[1]<0:
            print("BOUNCEYNEG")
            self.vel[1]*=-1
            self.pos[1]=0
        
    
    def tick(self):
        if self.timer!=-1:
            self.timer-=1
            
    # returns true on death:
    def applyDamage(self, dmg):
        self.health-=dmg
        return self.health<0
    
    def getDict(self):
        d={}
        d['pos']=self.pos
        d['vel']=self.vel
        d['health']=self.health
        d['inventory']=self.inventory
        d['type']=self.type
        d['dims']=[self.width, self.height]
        d['timer']=self.timer
        
        return d

class BattleTanks:
    def __init__(self, state=None, remotePlayers=None):
        self.ais=remotePlayers
        self.items=[]
        self.projectiles=[]
        self.timedProjectiles=[]
        self.deathList=[]
        # Represents sprites in the JS engine:
        self.graphics=[]
        self.dims=(WIDTH, HEIGHT)
        self.objectSet={}
        self.tickct=0
        # This is only kept for compatibility with the new ref.
        self.state={}
        
        # Populate the world with players
        for ai in self.ais:
            pos=[random.randint(0, self.dims[0]-1), random.randint(0, self.dims[1]-1)]
            self.objectSet[ai]=BTObject(ai, pos, [0, 0], 100, "player", {}, 32, 32)
        
        # Populate the world with an initial item set (more will be added at random based on tick):
        for i in range(4):
            itemID=str(uuid.uuid4())
            pos=[random.randint(0, self.dims[0]-1), random.randint(0, self.dims[1]-1)]
            self.items.append(itemID)
            self.objectSet[itemID]=BTObject(itemID, pos, [0, 0], 1, "item", [ITEMS[i]], 32, 32)
            
            
    def getStateForPlayer(self, player):
        ps={}
        # This state is going to be large and slow (for now)
        ps['id']=player
        ps['boarddims']=[WIDTH, HEIGHT]
        ps['players']={p: self.objectSet[p].getDict() for p in self.ais}
        ps['items']={p: self.objectSet[p].getDict() for p in self.items}
        ps['projectiles']={p: self.objectSet[p].getDict() for p in self.projectiles}
        ps['timedprojectiles']={p: self.objectSet[p].getDict() for p in self.timedProjectiles}
        ps['graphics']={p: self.objectSet[p].getDict() for p in self.graphics}
        return ps
    
    def doPickup(self, aiid, itid, aiobj, itobj):
        # Add item attributes to player:
        for it in itobj.inventory:
            # Right now, we're only going to consider health:
            health=it['health']
            self.objectSet[aiid].health+=health
        # Item object cleanup is performed externally.ooo
    
    # Allow player input:
    def makeMove(self, move, playerID=None):
        if playerID is None:
            return
        
        ai=playerID
        
        if ai in self.deathList:
            # Don't allow dead AIs to submit moves.
            return
        
        # Move and stopmoving are mutually exclusive:
        if "move" in move:
            angle=move['move']
            xdir=math.cos(angle)
            ydir=math.sin(angle)
            
            # scale the vector to the correct velocity:
            dst=euDist((xdir, ydir), (0, 0))
            sf=MOVESPEED/dst
            xdir*=sf
            ydir*=sf
            print("\tSETTING VELOCITY TO ", [xdir, ydir])
            self.objectSet[ai].vel=[xdir, ydir]
            
        elif "stopmoving" in move:
            # Zero out velocity:
            self.objectSet[ai].vel=[0, 0]
        
        # fire and place (planting a bomb) are mutually exclusive)
        if "fire" in move:
            itemID=str(uuid.uuid4())
            pos=self.objectSet[ai].pos
            
            angle=move['fire']
            xdir=math.cos(angle)
            ydir=math.sin(angle)
            
            # scale the vector to the correct velocity:
            dst=euDist((xdir, ydir), [0, 0])
            sf=PROJMOVESPEED/dst
            xdir*=sf
            ydir*=sf
            self.projectiles.append(itemID)
            self.objectSet[itemID]=BTObject(itemID, pos, [xdir, ydir], 1, "projectile", [], 32, 32)
        
        elif "place" in move:
            # todo: implement this
            pass
        
    def validateMove(self, move, playerID=None):
        ## This is a stub at the moment:
        return True
        
        
    def tick(self):
        self.tickct+=1
        
        if len(self.ais)==0 and len(self.deathList)!=0:
            # The winner is the last AI to die:
            print("\tCALLING MATCH WITH DEATHLIST: ", self.deathList)
            self.state['Winner']=self.deathList[-1]
            return
        
        elif len(self.ais)==0:
            print("\tCALLING MATCH WITH EMPTY DEATHLIST: ", self.deathList)
            self.state['Winner']=None
            return
        
        rawobjlist=[]
        
        # Perform object movement updates:
        for o in self.objectSet:
            self.objectSet[o].move()
            self.objectSet[o].tick()
            
        # Determine if any graphics have expired:
        graphicremlist=[]
        for g in self.graphics:
            gobj=self.objectSet[g]
            if gobj.timer==0:
                graphicremlist.append(g)
        
        # Remove all expired graphics prior to computing explosions, projectiles, etc:
        graphicremlist=list(set(graphicremlist))
        for g in graphicremlist:
            del self.graphics[self.graphics.index(g)]
            del self.objectSet[g]
            
        rawobjlist=[self.objectSet[o] for o in self.objectSet]
        
        # Determine if any timed projectiles should go off:
        objremlist=[]
        for t in self.timedProjectiles:
            tobj=self.objectSet[t]
            if tobj.timer==0:
                objremlist.append(t)
                splashlist=tobj.splash(rawobjlist, BOMBDAMAGE, BOMBSCLFAC, BOMBSPLASH)
                
                for s in splashlist:
                    # We're already removing the current object.
                    sobj=s[0]
                    sdmg=s[1]
                    if self.objectSet[sobj].applyDamage(sdmg):
                        objremlist.append(sobj)
                
                
                
        # Perform collision detection:
        for ai in self.ais:
            aiobj=self.objectSet[ai]
            
            # Perform item collision detection and pickups
            for i in self.items:
                it=self.objectSet[i]
                if aiobj.detectHit(it):
                    self.doPickup(ai, i, aiobj, it)
                    objremlist.append(i)
            
            # Perform projectile collision detection and splash damage:
            for p in self.projectiles:
                pobj=self.objectSet[p]
                
                if aiobj.detectHit(pobj):
                    splashlist=pobj.splash(rawobjlist, PROJECTILEDAMAGE, PROJECTILESCLFAC, PROJECTILESPLASH)
                    for s in splashlist:
                        sobj=s[0]
                        sdmg=s[1]
                        if self.objectSet[sobj].applyDamage(sdmg):
                            objremlist.append(sobj)
        
        # De-duplicate the object removal list:
        objremlist=list(set(objremlist))
        
        for o in objremlist:
            obj=self.objectSet[o]
            
            # Since all graphics have already been culled as appropriate, let's spawn an explosion graphic:
            explID=str(uuid.uuid4())
            pos=obj.pos
            self.graphics.append(explID)
            self.objectSet[explID]=BTObject(explID, pos, [0, 0], 1, "explosion", [], 32, 32, timer=7)
            
            if obj.type=="player":
                if obj.id not in self.deathList:
                    self.deathList.append(obj)
                    del self.ais[self.ais.index(o)]
                    
                    # Then we're the de facto winner:
                    if len(self.ais)==0:
                        self.state['Winner']=obj.id
                        print("\tCALLING MATCH WITH WINNER: ", self.state['Winner'])
                        return
            elif obj.type=="item":
                del self.items[self.items.index(o)]
                del self.objectSet[o]
            elif obj.type=="projectile":
                del self.projectiles[self.projectiles.index(o)]
                del self.objectSet[o]
            elif obj.type=="timedprojectile":
                del self.timedProjectilesp[self.timedProjectiles.index(o)]
                del self.objectSet[o]
            elif obj.type=="explosion":
                pass # Don't allow explosions to get destroyed by collisions.
            else:
                del self.objectSet[o]
                
                    
        
        
